"""
User Profile Backup Tool
This program searches for the most immediate convenient destination for data backup,
and if a larger destination is detected, that drive letter is chosen. This works for
secondary internal hard drives, as well as external hard drives and backs up all 
detected files and folders of the most common directories of the current logged in user.
Written by Chad Jordan
3/16/2019
"""
import os
import shutil
import time
import datetime
import tkinter as tk
from tkinter import ttk, messagebox
import threading
import subprocess
from pathlib import Path

class BackupApp:
    def __init__(self, root):
        self.root = root
        self.root.title("User Profile Backup Tool")
        self.root.geometry("700x600")
        self.root.resizable(True, True)
        self.username = os.environ.get('USERNAME')
        self.user_profile = os.environ.get('USERPROFILE')
        self.date = datetime.datetime.now().strftime("%Y-%m-%d")
        self.backup_root = "UserBackups"
        self.backup_destination = f"C:\\{self.backup_root}\\{self.username}\\{self.date}"
        self.folders_to_backup = ["Desktop", "Documents", "Downloads", "Pictures", "Music", "Videos"]
        self.browsers = ["Google Chrome", "Mozilla Firefox", "Microsoft Edge", "Opera"]
        self.total_source_size = 0
        self.current_backup_size = 0
        self.backup_running = False
        self.start_time = None
        self.stop_requested = False
        self.current_process_id = None
    
        #Set up cleanup on window close
        self.root.protocol("WM_DELETE_WINDOW", self.confirm_exit)
    
        #Create UI
        self.create_ui()
    
    def create_ui(self):
        #Title label
        title_label = ttk.Label(self.root, text="User Profile Backup Tool", font=("Arial", 14, "bold"))
        title_label.pack(pady=(15, 5))
        
        #Description label
        desc_label = ttk.Label(self.root, text=f"Backing up user profile for: {self.username}")
        desc_label.pack(pady=(0, 15))
        
        #Log box
        self.log_frame = ttk.LabelFrame(self.root, text="Backup Log")
        self.log_frame.pack(fill=tk.BOTH, expand=True, padx=20, pady=10)
        
        self.log_box = tk.Text(self.log_frame, height=15, width=80)
        self.log_box.pack(side=tk.LEFT, fill=tk.BOTH, expand=True, padx=5, pady=5)
        
        self.scrollbar = ttk.Scrollbar(self.log_frame, command=self.log_box.yview)
        self.scrollbar.pack(side=tk.RIGHT, fill=tk.Y, pady=5)
        self.log_box.config(yscrollcommand=self.scrollbar.set)
        self.log_box.config(state=tk.DISABLED)
        
        #Progress frame
        self.progress_frame = ttk.LabelFrame(self.root, text="Backup Progress")
        self.progress_frame.pack(fill=tk.X, padx=20, pady=10)
        
        #Progress bar
        self.progress_bar = ttk.Progressbar(self.progress_frame, orient=tk.HORIZONTAL, length=660, mode='determinate')
        self.progress_bar.pack(fill=tk.X, padx=10, pady=(10, 5))
        
        #Status frame for labels
        self.status_frame = ttk.Frame(self.progress_frame)
        self.status_frame.pack(fill=tk.X, padx=10, pady=(0, 10))
        
        #Create a grid layout for status information
        self.status_frame.columnconfigure(0, weight=1)
        self.status_frame.columnconfigure(1, weight=1)
        
        #Percentage label
        self.percentage_label = ttk.Label(self.status_frame, text="Percentage Complete: 0%")
        self.percentage_label.grid(row=0, column=0, sticky=tk.W, pady=2)
        
        #Time elapsed label
        self.time_elapsed_label = ttk.Label(self.status_frame, text="Time Elapsed: 00:00:00")
        self.time_elapsed_label.grid(row=0, column=1, sticky=tk.E, pady=2)
        
        #Data copied label
        self.data_copied_label = ttk.Label(self.status_frame, text="Data Copied: 0 MB")
        self.data_copied_label.grid(row=1, column=0, sticky=tk.W, pady=2)
        
        #Estimated time remaining label
        self.time_remaining_label = ttk.Label(self.status_frame, text="Estimated Time Remaining: Calculating...")
        self.time_remaining_label.grid(row=1, column=1, sticky=tk.E, pady=2)
        
        #Buttons frame
        self.button_frame = ttk.Frame(self.root)
        self.button_frame.pack(fill=tk.X, padx=20, pady=(5, 30)) 
        
        #Start button
        self.start_button = ttk.Button(self.button_frame, text="Start Backup", command=self.start_backup, width=20)
        self.start_button.pack(side=tk.LEFT, padx=(0, 10))
        
        #Stop button
        self.stop_button = ttk.Button(self.button_frame, text="Stop Backup", command=self.stop_backup, width=20, state=tk.DISABLED)
        self.stop_button.pack(side=tk.LEFT)
        
        #Exit button
        self.exit_button = ttk.Button(self.button_frame, text="Exit", command=self.confirm_exit, width=15)
        self.exit_button.pack(side=tk.RIGHT)
        
        #Status bar
        self.status_bar = ttk.Label(self.root, text="Ready", relief=tk.SUNKEN, anchor=tk.W)
        self.status_bar.pack(side=tk.BOTTOM, fill=tk.X)
        
        #Set up a timer for updating elapsed time
        self.update_elapsed_time()
    
    def write_log(self, message):
        self.log_box.config(state=tk.NORMAL)
        timestamp = datetime.datetime.now().strftime("%H:%M:%S")
        self.log_box.insert(tk.END, f"{timestamp} - {message}\n")
        self.log_box.see(tk.END)
        self.log_box.config(state=tk.DISABLED)
        self.root.update_idletasks()
    
    def update_progress(self, percentage, current_size=None):
        self.progress_bar['value'] = percentage
        self.percentage_label.config(text=f"Percentage Complete: {percentage:.1f}%")
        
        if current_size is not None:
            #Update data copied
            if current_size < 1024**3:
                self.data_copied_label.config(text=f"Data Copied: {current_size / (1024**2):.2f} MB")
            else:
                self.data_copied_label.config(text=f"Data Copied: {current_size / (1024**3):.2f} GB")
            
            #Update estimated time remaining
            if self.start_time and percentage > 0:
                elapsed_seconds = time.time() - self.start_time
                total_estimated_seconds = (elapsed_seconds / percentage) * 100
                remaining_seconds = total_estimated_seconds - elapsed_seconds
                
                if remaining_seconds > 0:
                    hours, remainder = divmod(int(remaining_seconds), 3600)
                    minutes, seconds = divmod(remainder, 60)
                    self.time_remaining_label.config(text=f"Estimated Time Remaining: {hours:02}:{minutes:02}:{seconds:02}")
                else:
                    self.time_remaining_label.config(text="Estimated Time Remaining: Almost done")
        
        self.root.update_idletasks()
    
    def update_elapsed_time(self):
        if self.start_time and self.backup_running:
            elapsed_seconds = int(time.time() - self.start_time)
            hours, remainder = divmod(elapsed_seconds, 3600)
            minutes, seconds = divmod(remainder, 60)
            self.time_elapsed_label.config(text=f"Time Elapsed: {hours:02}:{minutes:02}:{seconds:02}")
        
        #Schedule the next update in 1 second
        self.root.after(1000, self.update_elapsed_time)
    
    def check_drives(self):
        self.write_log("Checking for Removable/External media for backup...")
        best_drive = None
        max_space = 0
        
        #Get all drives except C:
        for drive in [f"{d}:\\" for d in "DEFGHIJKLMNOPQRSTUVWXYZ" if os.path.exists(f"{d}:")]:
            try:
                total, used, free = shutil.disk_usage(drive)
                free_gb = free / (1024**3)
                self.write_log(f"Drive {drive}: {free_gb:.2f} GB free")
                
                if free_gb > max_space:
                    max_space = free_gb
                    best_drive = drive[0]
            except:
                pass
        
        if best_drive:
            self.backup_destination = f"{best_drive}:\\{self.backup_root}\\{self.username}\\{self.date}"
            self.write_log(f"Found a better location on Drive {best_drive}:. Backing up to this drive.")
        else:
            self.backup_destination = f"C:\\{self.backup_root}\\{self.username}\\{self.date}"
            self.write_log("No external drives found with sufficient space. Using default location on C: Drive.")
        
        os.makedirs(self.backup_destination, exist_ok=True)
        self.write_log(f"Backup Destination set to: {self.backup_destination}")
    
    def calculate_total_size(self):
        self.write_log("Calculating total size of files to backup...")
        total_size = 0
        
        for folder in self.folders_to_backup:
            source_path = os.path.join(self.user_profile, folder)
            if os.path.exists(source_path):
                try:
                    #Simplified size calculation
                    folder_size = 0
                    for dirpath, _, filenames in os.walk(source_path):
                        for filename in filenames:
                            try:
                                file_path = os.path.join(dirpath, filename)
                                folder_size += os.path.getsize(file_path)
                            except:
                                pass
                    
                    if folder_size > 0:
                        total_size += folder_size
                        self.write_log(f"Size of {folder} folder: {folder_size / (1024**3):.2f} GB")
                except Exception as e:
                    self.write_log(f"Error calculating size of {folder} folder: {str(e)}")
        
        #Add buffer for browser bookmarks
        total_size += 10 * 1024 * 1024 
        
        #Ensure we have a minimum size to avoid division by zero
        if total_size < 1024 * 1024: 
            total_size = 1024 * 1024 
        
        self.write_log(f"Total size to backup: {total_size / (1024**3):.2f} GB")
        
        return total_size
    
    def check_progress(self):
        last_size = 0
        stalled_count = 0
        last_folder_check_time = time.time()
        
        while self.backup_running and not self.stop_requested:
            try:
                #Calculate current backup size more efficiently
                current_time = time.time()
                current_size = 0
                
                #Only do a full recalculation every 10 seconds to reduce disk I/O
                if current_time - last_folder_check_time > 10:
                    try:
                        total_size_cmd = f'dir /s /a "{self.backup_destination}" | findstr "File(s)"'
                        result = subprocess.run(total_size_cmd, shell=True, capture_output=True, text=True, timeout=5)
                        output_lines = result.stdout.strip().split('\n')
                        for line in output_lines:
                            if "bytes" in line:
                                parts = line.strip().split()
                                for i, part in enumerate(parts):
                                    if part == "bytes":
                                        try:
                                            size_str = parts[i-1].replace(',', '')
                                            current_size = int(size_str)
                                            break
                                        except:
                                            pass
                        
                        last_folder_check_time = current_time
                    except Exception as e:
                        #Fall back to previous method
                        try:
                            current_size = sum(os.path.getsize(os.path.join(dirpath, filename)) 
                                            for dirpath, _, filenames in os.walk(self.backup_destination) 
                                            for filename in filenames)
                        except:
                            current_size = last_size
                else:
                    #Use the last known size between full calculations
                    current_size = self.current_backup_size
                
                self.current_backup_size = current_size
                
                #Check if backup is stalled
                if current_size == last_size:
                    stalled_count += 1
                else:
                    stalled_count = 0
                    last_size = current_size
                
                if stalled_count >= 15:
                    self.write_log("Warning: Backup appears to be stalled. Continuing to monitor...")
                    stalled_count = 10 
                
                if self.total_source_size > 0:
                    percentage = min(99, max(1, (current_size / self.total_source_size) * 100))
                    self.update_progress(percentage, current_size)
                    if hasattr(self, 'last_logged_percentage') and hasattr(self, 'last_progress_time'):
                        if (abs(percentage - self.last_logged_percentage) >= 10 or 
                            (current_time - self.last_progress_time) >= 120):
                            self.write_log(f"Progress: {current_size / (1024**3):.2f} GB copied ({percentage:.1f}%)")
                            self.last_logged_percentage = percentage
                            self.last_progress_time = current_time
                    else:
                        self.last_logged_percentage = percentage
                        self.last_progress_time = current_time
            except Exception as e:
                pass  #Silently handle errors to reduce log spam
            
            time.sleep(2)
    
    def backup_folder(self, source, destination):
        if self.stop_requested:
            return False
            
        folder_name = os.path.basename(source)
        self.write_log(f"Backing up {folder_name}...")
        self.status_bar.config(text=f"Backing up: {folder_name}")
        
        #Create destination if it doesn't exist
        os.makedirs(destination, exist_ok=True)
        
        try:
            #Create a list of problematic paths to exclude
            exclude_dirs = [
                "AppData\\Local\\Temp", 
                "AppData\\LocalLow", 
                "AppData\\Local\\Microsoft\\Windows\\INetCache",
                "AppData\\Local\\Microsoft\\Windows\\WER",
                "AppData\\Local\\Microsoft\\Windows\\WebCache",
                "AppData\\Local\\CrashDumps",
                "AppData\\Local\\Microsoft\\Windows\\Explorer"
            ]
            
            #Build exclude parameters
            exclude_params = ""
            for exclude_dir in exclude_dirs:
                full_path = os.path.join(self.user_profile, exclude_dir)
                if os.path.exists(full_path):
                    exclude_params += f' /XD "{full_path}"'
            
            #Enhanced robocopy command
            cmd = (
                f'robocopy "{source}" "{destination}" /E /Z /XJ /XA:SH {exclude_params} '
                f'/R:0 /W:0 /MT:8 /NP /NFL /NDL /NC /NS /LOG+:"{self.backup_destination}\\robocopy_{folder_name}.txt"'
            )
            
            self.write_log(f"Starting robocopy for {folder_name}...")
            
            process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            
            #Set a timeout for each folder (20 minutes for large folders)
            start_time = time.time()
            timeout = 1200  #20 minutes
            
            while process.poll() is None:
                if self.stop_requested:
                    process.terminate()
                    self.kill_robocopy_processes()
                    self.write_log(f"Backup of {folder_name} was interrupted.")
                    return False
                
                if time.time() - start_time > timeout:
                    self.write_log(f"Timeout reached for {folder_name}. Moving to next folder.")
                    process.terminate()
                    self.kill_robocopy_processes()
                    return False
                    
                time.sleep(0.5)
            
            return_code = process.returncode
            
            #Robocopy return codes:
            #0 = No errors, no copying
            #1 = Files copied successfully
            #2 = Extra files/dirs detected but no failures
            #3 = Some files/dirs couldn't be copied (errors or busy files)
            #4+ = Serious error
            if return_code >= 0 and return_code <= 3:
                self.write_log(f"{folder_name} backup completed with return code {return_code}.")
                
                #Double-check if we need to retry with different options for large folders
                if folder_name in ["Desktop", "Downloads", "Documents"] and return_code == 3:
                    self.write_log(f"Some files in {folder_name} couldn't be copied. Trying again with different options...")
                    
                    #Try again for problematic files
                    retry_cmd = (
                        f'robocopy "{source}" "{destination}" /E /Z /XJ /XA:SH {exclude_params} '
                        f'/R:0 /W:0 /MT:4 /NP /NFL /NDL /NC /NS /LOG+:"{self.backup_destination}\\robocopy_{folder_name}_retry.txt"'
                    )
                    
                    subprocess.run(retry_cmd, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                    self.write_log(f"Retry of {folder_name} completed.")
                
                return True
            else:
                self.write_log(f"Error backing up {folder_name}. Robocopy returned code {return_code}.")
                return False
                
        except Exception as e:
            self.write_log(f"Error backing up {folder_name}: {str(e)}")
            return False
    
    def backup_browser_bookmarks(self):
        if self.stop_requested:
            return
        
        self.write_log("Backing up browser bookmarks...")
        self.status_bar.config(text="Backing up: Browser Bookmarks")
        
        #Create the browser bookmarks directory
        bookmarks_dir = os.path.join(self.backup_destination, "BrowserBookmarks")
        os.makedirs(bookmarks_dir, exist_ok=True)
        
        #Google Chrome - multiple locations to ensure we find it
        chrome_locations = [
            os.path.join(self.user_profile, "AppData", "Local", "Google", "Chrome", "User Data", "Default"),
            os.path.join(self.user_profile, "AppData", "Local", "Google", "Chrome", "User Data", "Profile 1"),
            os.path.join(self.user_profile, "AppData", "Local", "Google", "Chrome", "User Data", "Profile 2")
        ]
        
        chrome_files = ["Bookmarks", "Bookmarks.bak", "Favicons"]
        chrome_backed_up = False
        
        for chrome_path in chrome_locations:
            if os.path.exists(chrome_path):
                chrome_backup_dir = os.path.join(bookmarks_dir, "Chrome", os.path.basename(chrome_path))
                os.makedirs(chrome_backup_dir, exist_ok=True)
                
                for file in chrome_files:
                    file_path = os.path.join(chrome_path, file)
                    if os.path.exists(file_path):
                        try:
                            shutil.copy2(file_path, chrome_backup_dir)
                            chrome_backed_up = True
                            self.write_log(f"Chrome {file} backed up from {os.path.basename(chrome_path)}.")
                        except Exception as e:
                            self.write_log(f"Error backing up Chrome {file}: {str(e)}")
        
        if chrome_backed_up:
            self.write_log("Google Chrome bookmarks backed up successfully.")
        else:
            self.write_log("Could not find Google Chrome bookmarks.")
        
        #Check Firefox profiles
        firefox_profiles_dir = os.path.join(self.user_profile, "AppData", "Roaming", "Mozilla", "Firefox", "Profiles")
        firefox_backed_up = False
        
        if os.path.exists(firefox_profiles_dir):
            for profile in os.listdir(firefox_profiles_dir):
                profile_dir = os.path.join(firefox_profiles_dir, profile)
                if os.path.isdir(profile_dir):
                    firefox_files = ["places.sqlite", "favicons.sqlite", "bookmarkbackups"]
                    firefox_backup_dir = os.path.join(bookmarks_dir, "Firefox", profile)
                    os.makedirs(firefox_backup_dir, exist_ok=True)
                    
                    for file in firefox_files:
                        file_path = os.path.join(profile_dir, file)
                        if os.path.exists(file_path):
                            try:
                                if os.path.isdir(file_path):
                                    # For directories like bookmarkbackups
                                    shutil.copytree(file_path, os.path.join(firefox_backup_dir, file), dirs_exist_ok=True)
                                else:
                                    shutil.copy2(file_path, firefox_backup_dir)
                                firefox_backed_up = True
                                self.write_log(f"Firefox {file} backed up from profile {profile}.")
                            except Exception as e:
                                self.write_log(f"Error backing up Firefox {file}: {str(e)}")
        
        if firefox_backed_up:
            self.write_log("Mozilla Firefox bookmarks backed up successfully.")
        else:
            self.write_log("Could not find Mozilla Firefox bookmarks.")
        
        #Microsoft Edge - multiple possible locations
        edge_locations = [
            os.path.join(self.user_profile, "AppData", "Local", "Microsoft", "Edge", "User Data", "Default"),
            os.path.join(self.user_profile, "AppData", "Local", "Microsoft", "Edge", "User Data", "Profile 1"),
            os.path.join(self.user_profile, "AppData", "Local", "Packages", "Microsoft.MicrosoftEdge_8wekyb3d8bbwe", "AC", "MicrosoftEdge", "User", "Default")
        ]
        
        edge_files = ["Bookmarks", "Bookmarks.bak", "Favicons"]
        edge_backed_up = False
        
        for edge_path in edge_locations:
            if os.path.exists(edge_path):
                edge_backup_dir = os.path.join(bookmarks_dir, "Edge", os.path.basename(edge_path))
                os.makedirs(edge_backup_dir, exist_ok=True)
                
                #Check for favorites folder (old Edge)
                favorites_dir = os.path.join(edge_path, "Favorites")
                if os.path.exists(favorites_dir):
                    try:
                        shutil.copytree(favorites_dir, os.path.join(edge_backup_dir, "Favorites"), dirs_exist_ok=True)
                        edge_backed_up = True
                        self.write_log("Edge favorites folder backed up.")
                    except Exception as e:
                        self.write_log(f"Error backing up Edge favorites folder: {str(e)}")
                
                #Check for bookmark files (new Edge)
                for file in edge_files:
                    file_path = os.path.join(edge_path, file)
                    if os.path.exists(file_path):
                        try:
                            shutil.copy2(file_path, edge_backup_dir)
                            edge_backed_up = True
                            self.write_log(f"Edge {file} backed up from {os.path.basename(edge_path)}.")
                        except Exception as e:
                            self.write_log(f"Error backing up Edge {file}: {str(e)}")
        
        if edge_backed_up:
            self.write_log("Microsoft Edge bookmarks backed up successfully.")
        else:
            self.write_log("Could not find Microsoft Edge bookmarks.")
        
        self.write_log("Browser bookmarks backup completed.")

        #Opera
        opera_dir = os.path.join(self.user_profile, "AppData", "Roaming", "Opera Software", "Opera Stable")
        if os.path.exists(opera_dir):
            opera_bookmarks_file = os.path.join(opera_dir, "Bookmarks")
            if os.path.exists(opera_bookmarks_file):
                opera_backup_dir = os.path.join(self.backup_destination, "BrowserBookmarks", "Opera")
                os.makedirs(opera_backup_dir, exist_ok=True)
                shutil.copy2(opera_bookmarks_file, opera_backup_dir)
                self.write_log("Opera bookmarks backed up.")
    
    def stop_backup(self):
        """Stop the backup process"""
        if not self.backup_running:
            return
    
        result = messagebox.askyesno("Confirm Stop", "Are you sure you want to stop the backup process?")
        if result:
            self.write_log("User requested to stop the backup process.")
            self.stop_requested = True
            self.status_bar.config(text="Stopping backup... Please wait.")
        
            #Disable the stop button while we're stopping
            self.stop_button.config(state=tk.DISABLED)
        
            #Kill any running robocopy processes
            self.kill_robocopy_processes()
        
            #Wait a moment for processes to terminate
            self.root.after(1000, self.check_backup_stopped)

    def check_backup_stopped(self):
        """Check if backup has been fully stopped"""
        #Kill robocopy processes
        self.kill_robocopy_processes()
    
        #Check if backup is still running after 2 seconds
        if self.backup_running:
            #Set a maximum number of checks to avoid infinite loop
            if not hasattr(self, 'stop_check_count'):
                self.stop_check_count = 0
        
            self.stop_check_count += 1
        
            if self.stop_check_count >= 5:
                self.backup_running = False
                self.start_button.config(state=tk.NORMAL)
                self.stop_button.config(state=tk.DISABLED)
                self.status_bar.config(text="Backup stopped forcefully.")
                self.write_log("Backup process was forcefully terminated.")
                return
        
            #Check in one second
            self.root.after(1000, self.check_backup_stopped)
        else:
            self.status_bar.config(text="Backup stopped.")
            self.stop_check_count = 0

    def kill_robocopy_processes(self):
        """Kill all robocopy processes"""
        try:
            #aggressive kill for robocopy
            subprocess.run("taskkill /f /im robocopy.exe", shell=True, 
                        stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            self.write_log("Terminated all robocopy processes.")
        except Exception as e:
            self.write_log(f"Error terminating robocopy processes: {str(e)}")

    def confirm_exit(self):
        """Confirm before exiting the application"""
        if self.backup_running:
            result = messagebox.askyesno("Backup in Progress", 
                                     "A backup is currently in progress. Are you sure you want to exit?")
            if not result:
                return
        
            self.stop_requested = True
            self.write_log("User requested to exit. Stopping backup process.")
            self.kill_robocopy_processes()
        
            #Wait a moment for processes to terminate
            self.root.after(500, self.force_exit)
        else:
            self.root.destroy()

    def force_exit(self):
        """Force exit after killing processes"""
        #Kill robocopy processes one more time to be sure
        self.kill_robocopy_processes()
        self.root.destroy()
        
    def start_backup(self):
        if self.backup_running:
            return
    
        self.backup_running = True
        self.stop_requested = False
        self.start_time = time.time()
    
        #Update button states
        self.start_button.config(state=tk.DISABLED)
        self.stop_button.config(state=tk.NORMAL)
    
        #Reset progress indicators
        self.progress_bar['value'] = 0
        self.percentage_label.config(text="Percentage Complete: 0%")
        self.time_elapsed_label.config(text="Time Elapsed: 00:00:00")
        self.time_remaining_label.config(text="Estimated Time Remaining: Calculating...")
        self.data_copied_label.config(text="Data Copied: 0 MB")
        self.status_bar.config(text="Backup in progress...")
    
        #Start the backup in a separate thread
        threading.Thread(target=self.run_backup, daemon=True).start()
    
        #Start progress monitoring in a separate thread
        threading.Thread(target=self.check_progress, daemon=True).start()
    
    def run_backup(self):
        try:
            self.check_drives()
        
            self.total_source_size = self.calculate_total_size()
            
            self.update_progress(1)
            
            self.write_log(f"Starting backup for user: {self.username}")
            self.write_log(f"Backup Destination: {self.backup_destination}")
            self.write_log("-" * 50)
            
            priority_folders = ["Desktop", "Documents", "Downloads", "Pictures", "Music", "Videos"]
            
            for folder in priority_folders:
                if self.stop_requested:
                    self.write_log("Backup process was stopped by user.")
                    break
                    
                source_path = os.path.join(self.user_profile, folder)
                destination_path = os.path.join(self.backup_destination, folder)
                
                if os.path.exists(source_path):
                    success = self.backup_folder(source_path, destination_path)
                    if not success:
                        self.write_log(f"Warning: Backup of {folder} may be incomplete. Continuing with next folder.")
                        
                        #For important folders, try one more times with different options
                        if folder in ["Desktop", "Documents", "Downloads"]:
                            self.write_log(f"Attempting one more backup of {folder} with different options...")
                            try:
                                #Simplified backup for retry
                                simple_cmd = (
                                    f'robocopy "{source_path}" "{destination_path}" /E /XJ /R:0 /W:0 /MT:2 '
                                    f'/XA:SH /NFL /NDL /NC /NS /LOG+:"{self.backup_destination}\\robocopy_{folder}_final.txt"'
                                )
                                subprocess.run(simple_cmd, shell=True, timeout=600)
                                self.write_log(f"Final attempt for {folder} completed.")
                            except Exception as e:
                                self.write_log(f"Error in final attempt for {folder}: {str(e)}")
            
            #Backup browser bookmarks if not stopped
            if not self.stop_requested:
                self.backup_browser_bookmarks()
                self.update_progress(100)
                self.time_remaining_label.config(text="Estimated Time Remaining: Complete")
                self.write_log("Backup completed successfully!")
                self.status_bar.config(text="Backup completed successfully!")
            else:
                self.write_log("Backup was not completed due to user interruption.")
                self.status_bar.config(text="Backup stopped by user.")
                
        except Exception as e:
            self.write_log(f"Error during backup: {str(e)}")
            self.status_bar.config(text="Backup failed with errors.")
        finally:
            self.backup_running = False
            self.start_button.config(state=tk.NORMAL)
            self.stop_button.config(state=tk.DISABLED)

if __name__ == "__main__":
    root = tk.Tk()
    app = BackupApp(root)
    root.mainloop()
